
//--- PdfObj.hpp
//--- Copyright Thilo Brai, 2012
//--- Creative Commons Lizenz: CC-by-sa
//
//--- Hier sind die Pdf-Objekte definiert:
//--- Number, String, Stream, Array und Dictionary.
//--- Auerdem einige Pdf-Funktionen wie:
//--- PdfDate (), Win2PdfEncoding () etc.
//
//--- Wessen Texteditor andere Newlines nutzt, kann 
//--- hier z. B. NL = "\n" setzten.

#ifndef PdfObj_hpp
#define PdfObj_hpp

#include <math.h>
#include "cString.hpp"

//--- neue Zeile
const char *NL = "\r\n";


enum ObjectType {
	ObjNull = 0,
	ObjNumber,
	ObjString,
	ObjArray,
	ObjStream,
	ObjDict
};



class cObject {
public:
	cObject () { saveCounter = 0; }
	~cObject () { return; }

	int saveCounter;

	virtual ObjectType  getType ()   { return ObjNull; }
	virtual int         getObjNum () { return -1; }
	virtual const char *getObjString (cAString &ret) { // liefet alles auer "1 0 obj" und "endobj"
		return "null";
	}
	virtual double getNumber () { return 0; }
};



class cNumber : public cObject {
public:
	cNumber ()         { number = 0; }
	cNumber (int i)    { number = i; }
	cNumber (double d) { number = d; }
	~cNumber () { return; }

	double number;

	ObjectType  getType () { return ObjNumber; }
	const char *getObjString (cAString &ret) {
		if (floor (number) == ceil (number))
			ret.Format ("%d", int (number));
		else
			ret.Format ("%f", number);
		return ret;
	}
	double getNumber () { return number; }
};



class cPdfString : public cObject {
public:
	cPdfString () { return; }
	cPdfString (const char *str) { string = str; }
	~cPdfString () { return; }

	ObjectType  getType () { return ObjString; }

	const char *set (const char *newString) {
		string = newString;
		return string;
	}

	const char *getObjString (cAString &ret) {
		return string;
	}

	virtual double getNumber () {
		double d = strtod (string, 0);
		return d; 
	}

	operator const char * () {
		return string;
	}
	const char * operator = (const char *str) {
		return set (str);
	}

public:
	cAString string;
};



//--- Bei PDF wird alles in Objects verpackt.
//--- Der Seiteninhalt steckt in Streams.
//--- Beim Speichern kriegt jeder Stream noch ein Dictionary 
//--- mit der /Length vorangestellt.
//--- 4 0 obj \n <</Length x>> stream \n ... \n endstream \n endobj
//---                                   ^bis^ = /Length
class cPdfStream : public cObject {
private:
	cPdfStream () { return; }
public:
	cPdfStream (int num) { objNum = num; }
	~cPdfStream () { return; }

	int         objNum;
	cAString stream;

	ObjectType  getType () { return ObjStream; }
	int         getObjNum () { return objNum; }

	void add (const char *string)	{
		stream += string;
		stream += NL;
	}

	const char *getObjString (cAString &ret) {	// liefet alles auer "1 0 obj" und "endobj"
		ret.Format ("<</Length %u>>%s", stream.Length (), NL);	// Wozu der Length-Key ntig ist, weis Adobe...
		cAString s1;
		s1 = "stream";
		s1 += NL;
		s1 += stream;
		s1 += "endstream";
		ret += s1;
		return ret;
	}
};


class cArray : public cObject {
	cArray () { return; }
public:
	cArray (int num) { objNum = num; }
	~cArray () {
		while (array.GetCount ()) {
			int last = array.GetCount () - 1;
			cObject *obj = array.GetIt (last);
			array.Remove (last);
			obj->saveCounter--;
			if (obj->saveCounter <= 0)
				delete obj;
		}
	}

	int                   objNum;
	tPtrArray <cObject *> array;

	ObjectType getType () { return ObjArray; }
	int        getObjNum () { return objNum; }

	int getLength () { return array.GetCount (); }

	/*cObject *operator [] (int index) {
		if ((index >= 0) && (index < (int)array.GetCount ()))
			return array.GetIt (index);
		return 0;
	}*/

	void add (cObject *obj) {
		array.Add (obj);
		obj->saveCounter++;
	}

	void push (cObject *obj) {
		add (obj);
	}

	cObject *get (int index) {
		return array.GetIt (index);
	}

	const char *getObjString (cAString &ret)	// liefet alles auer "1 0 obj" und "endobj"
	{
		int i, length = array.GetCount ();
		cAString temp;
		ret = "[";
		for (i = 0; i < length; i++) {
			ret += array.GetIt (i)->getObjString (temp);
			ret += " ";
		}
		ret += "]";
		return ret;
	}
};


class cPdfDict : public cObject {
	cPdfDict () { return; }
public:
	cPdfDict (int num, const char *type)
	{
		objNum = num;
		if (!type || !*type)
			type = "/dict";
		addKey ("/Type", type);	// z. B. Catalog, Pages, Page ...
	}
	~cPdfDict ()
	{
		while (values.GetCount ()) {
			names.RemoveAndDelete (names.GetCount () - 1);
			int last = values.GetCount () - 1;
			cObject *obj = values.GetIt (last);
			values.Remove (last);
			obj->saveCounter--;
			if (obj->saveCounter <= 0)
				delete obj;
		}
	}

	int                   objNum;
	cAStringArray         names;
	tPtrArray <cObject *> values;

	ObjectType  getType () { return ObjDict; }
	int         getObjNum () { return objNum; }

	int getLength () { return names.GetCount (); }

	void addKey (const char *name, cObject *value)
	{
		if (name && *name && value) {
			int i = findKeyIndex (name);
			if (i >= 0) {
				cObject **pValues = values.GetArray ();
				cObject *oldObj = pValues [i];
				if (oldObj != value) {
					pValues [i] = value;
					if (oldObj) {
						oldObj->saveCounter--;
						if (oldObj->saveCounter <= 0)
							delete oldObj;
					}
				}
				return;
			}
			names.AddString (name);
			values.Add (value);
			value->saveCounter++;
		}
	}

	void addKey (const char *name, cObject &value)
	{
		addKey (name, &value);
	}

	void addKey (const char *name, const char *text)
	{
		if (name && *name) {
			int i = findKeyIndex (name);
			if (i >= 0) {
				cObject **pValues = values.GetArray ();
				cObject *oldObj = pValues [i];
				pValues [i] = new cPdfString (text);
				pValues [i]->saveCounter++;
				if (oldObj) {
					oldObj->saveCounter--;
					if (oldObj->saveCounter <= 0)
						delete oldObj;
				}
				return;
			}
			names.AddString (name);
			cObject *obj = new cPdfString (text);
			values.Add (obj);
			obj->saveCounter++;
		}
	}

	void addKey (const char *name, int zahl)
	{
		cAString temp;
		temp.Format ("%d", zahl);
		addKey (name, temp);
	}

	cObject *getValueOfKey (const char *key)
	{
		int i = findKeyIndex (key);
		if (i >= 0)
			return values.GetIt (i);
		return 0;
	}

	const char *getKeyOfIndex (int index)
	{
		if (index < (int)names.GetCount ())
			return names.GetIt (index)->Get ();
		return 0;
	}

	cObject *getValueOfIndex (int index)
	{
		if (index < (int)values.GetCount ())
			return values.GetIt (index);
		return 0;
	}

	int findKeyIndex (const char *key)
	{
		int i, length = names.GetCount ();
		cAString **pNames = names.GetArray ();
		for (i = 0; i < length; i++)
			if (0 == strcmp (pNames [i]->Get (), key))
				return i;
		return -1;
	}

	const char *getObjString (cAString &ret) {
		ret = "<<";
		ret += NL;
		int i, length = names.GetCount ();
		for (i = 0; i < length; i++) {
			cAString temp, t2;
			//if ((typeof array [i + 1] == "string") | (typeof array [i + 1] == "number")) {
			temp = names.GetIt (i)->Get ();
			temp += " ";
			temp += values.GetIt (i)->getObjString (t2);
			temp += NL;
			//} else {
			//	r += array [i] + ' ' + array [i + 1] + " % " + typeof array [i + 1]
			//		+ NL;
			//}
			ret += temp;
		}
		ret += ">>";
		return ret;
	}
};


//--- Fr C++ noch eine Extra-Klasse fr Bookmarks, sonst
//--- bricht in PdfDoc::bookmarksMake () die Typecast-Hlle los
class cBookmark : public cObject {
public:
	cBookmark () { reset (); }
	cBookmark (const char *Title_, int Level_, int open_, const char *Dest_) { set (Title_, Level_, open_, Dest_); }
	~cBookmark () { return; }
	void reset () {
		Title.Reset ();
		Level = 0;
		open  = 0;
		Dest.Reset ();
		prev = 0;
	}
	void set (const char *Title_, int Level_, int open_, const char *Dest_) {
		Title.Set (Title_);
		Level = Level_;
		open  = open_;
		Dest.Set (Dest_);
	}
public:
	cAString Title;
	int Level;
	int open;
	cAString Dest;
	cPdfDict *prev;
};


//----------------------------------------
//----------------------------------------
//----- Einige Helper --------------------
//----------------------------------------
//----------------------------------------


//--- Datum und Zeit fr die Zeitangabe im Info-Dict
inline cAString PDFdate ()
{
	//return "(D:20120404181128+02'00')";
	cAString s;
	time_t sTime;

	time (&sTime);
	struct tm gmt = *gmtime (&sTime);
	struct tm loc = *localtime (&sTime);
	s.Format ("(D:%04u%02u%02u"	// Jahr Monat Tag
					"%02u%02u%02u"		// std min sec
					"%c"			// +/- zu gmt
					"%02u'00')"	// gmt-diff
					, 1900 + loc.tm_year, loc.tm_mon, loc.tm_mday
					, loc.tm_hour, loc.tm_min, loc.tm_sec
					, (gmt.tm_hour < loc.tm_hour) ? '+' : '-'
					, abs (gmt.tm_hour - loc.tm_hour)
					);
	return s;
}

//--- bersetzt vom WindowsEncoding ins PdfEncoding
//--- Bookmarks nutzen das PdfEncoding
inline cAString Win2PdfEncoding (const char *text)
{
	//signed char Win130 [] = {130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159};
	unsigned char Pdf130 [] = {145, 134, 140, 131, 129, 130,  26, 139, 151, 136, 150, 141, 153, 143, 144, 143, 144, 141, 142, 128, 133, 132,  31, 146, 157, 137, 156, 157, 158, 152};
	unsigned char ch;
	cAString s = text;
	int len = cAString::SaveStrlen (text);
	for (int i = 0; i < len; i++) {
		ch = ((unsigned char *)text) [i];
		if ((ch >= 130) && (ch <= 159))
			((unsigned char *)s.Get ()) [i] = Pdf130 [ch - 130];
	}
	return s;
}

inline cAString EscapePsText (const char *text)
{
	cAString s;
	int len = cAString::SaveStrlen (text);
	for (int i = 0; i < len; i++) {
		switch (text [i]) {
		case '\\': s += "\\\\"; break;
		case '(' : s += "\\(" ; break;
		case ')' : s += "\\)" ; break;
		case '\t': s += ' '   ; break;	// Whitespaces beachten
		case '\r': s += ' '   ; break;	// Whitespaces beachten
		case '\n': s += ' '   ; break;	// Whitespaces beachten
		default:
	#ifdef _D_EBUG
			if (text [i] & 0x80) {
				char octBuf [6];
				sprintf (octBuf, "\\%o", ((unsigned char *)text) [i]);
				s += octBuf;
			} else
	#endif
				s += text [i];
		}
	}
	return s;
}




//--- Ums Strings in Hex-kodierte Strings umzuwandeln
//--- Wandelt einen Paint-String zu einem Hex-String um.
//--- Leerzeichen werden ignoriert.
inline cAString pN1 (const char *s)
{
	int i, len;
	cAString temp = s;
	s = temp;
	char *p;
	while ((p = strchr (s, ' ')))
		strcpy (p, p + 1);

	if (strlen (s) & 0x07) {
		temp.Add ("........", 8 - (strlen (s) & 0x07));	// ein Punkt == wei
		s = temp;
	}

	len = strlen (s);

	cAString hex;
	for (i = 0; i < len; i += 8) {
		unsigned
			 byte = (s [i + 0] == '.' ? 0x80 : 0)
			|       (s [i + 1] == '.' ? 0x40 : 0)
			|       (s [i + 2] == '.' ? 0x20 : 0)
			|       (s [i + 3] == '.' ? 0x10 : 0)
			|       (s [i + 4] == '.' ? 0x08 : 0)
			|       (s [i + 5] == '.' ? 0x04 : 0)
			|       (s [i + 6] == '.' ? 0x02 : 0)
			|       (s [i + 7] == '.' ? 0x01 : 0);
		hex.FormatAdd ("%02x", byte);
	}
	hex += '>';
	return hex;
}


inline cAString roundmax (double d, int dezi)
{
	cAString s, format;
	format.Format ("%%.%df", dezi);
	s.Format (format, d);
	if (dezi) {
		while (s [s.Length () - 1] == '0')
			*strrchr (s, '0') = 0;
		if (s [s.Length () - 1] == '.')
			*strrchr (s, '.') = 0;
	}
	return s;
}

inline double mm2pt (double mm)
{
	return mm / 0.3527;
}

inline cAString circle (double centerX, double centerY, double radiusX, double radiusY)
{	/*
    100       0 m
	 100      -55,22852    55,22852 -100           0 -100 c
	-55,22852 -100        -100      -55,22852   -100    0 c
	-100       55,22852   -55,22852  100           0  100 c
	 55,22852  100         100       55,22852    100    0 c  	s 	*/
	cAString s;
	double r_5522x = radiusX * .5522852;	// 0,5522852
	double r_5522y = radiusY * .5522852;
	s.Format ("%s %s m%s"
				"%s %s %s %s %s %s c%s"
				"%s %s %s %s %s %s c%s"
				"%s %s %s %s %s %s c%s"
				"%s %s %s %s %s %s c%s"
				, roundmax ( centerX + radiusX, 2).Get (), roundmax ( centerY +      0, 2).Get (), NL	// m

				, roundmax ( centerX + radiusX, 2).Get (), roundmax ( centerY - r_5522y, 2).Get ()
				, roundmax ( centerX + r_5522x, 2).Get (), roundmax ( centerY - radiusY, 2).Get ()
				, roundmax ( centerX +       0, 2).Get (), roundmax ( centerY - radiusY, 2).Get (), NL	// c

				, roundmax ( centerX - r_5522x, 2).Get (), roundmax ( centerY - radiusY, 2).Get ()
				, roundmax ( centerX - radiusX, 2).Get (), roundmax ( centerY - r_5522y, 2).Get ()
				, roundmax ( centerX - radiusX, 2).Get (), roundmax ( centerY +       0, 2).Get (), NL	// c

				, roundmax ( centerX - radiusX, 2).Get (), roundmax ( centerY + r_5522y, 2).Get ()
				, roundmax ( centerX - r_5522x, 2).Get (), roundmax ( centerY + radiusY, 2).Get ()
				, roundmax ( centerX +       0, 2).Get (), roundmax ( centerY + radiusY, 2).Get (), NL	// c

				, roundmax ( centerX + r_5522x, 2).Get (), roundmax ( centerY + radiusY, 2).Get ()
				, roundmax ( centerX + radiusX, 2).Get (), roundmax ( centerY + r_5522y, 2).Get ()
				, roundmax ( centerX + radiusX, 2).Get (), roundmax ( centerY +       0, 2).Get (), NL	// c
				);
	return s;
}

inline cAString circle (double centerX, double centerY, double radius)
{
	return circle (centerX, centerY, radius, radius);
}


#endif // Obj_hpp
